package com.mobileconnected.amap.map;

import android.Manifest;
import android.annotation.SuppressLint;
import android.app.Activity;
import android.app.Application;
import android.content.Context;
import android.content.pm.PackageManager;
import android.graphics.Point;
import android.os.Bundle;
import android.util.Log;
import android.view.View;

import com.amap.api.maps.*;
import com.amap.api.maps.model.*;

import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.PluginRegistry;
import io.flutter.plugin.platform.PlatformView;

import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

import static com.mobileconnected.amap.map.MapPlugin.CREATED;
import static com.mobileconnected.amap.map.MapPlugin.DESTROYED;
import static com.mobileconnected.amap.map.MapPlugin.PAUSED;
import static com.mobileconnected.amap.map.MapPlugin.RESUMED;
import static com.mobileconnected.amap.map.MapPlugin.STARTED;
import static com.mobileconnected.amap.map.MapPlugin.STOPPED;

public class AmapMapController implements MethodChannel.MethodCallHandler,
        AMap.OnCameraChangeListener,
        AMap.OnInfoWindowClickListener,
        AMap.OnMarkerClickListener,
        AMap.OnMarkerDragListener,
        AMap.OnPolylineClickListener,
        AMap.OnMapClickListener,
        AMap.OnMapLongClickListener,
        Application.ActivityLifecycleCallbacks,
        AmapMapOptionsSink,
        PlatformView {
    private static final String TAG = "GoogleMapController";
    private final int id;
    private final AtomicInteger activityState;
    private final MethodChannel methodChannel;
    private final PluginRegistry.Registrar registrar;
    private final MapView mapView;
    private AMap amapMap;
    private boolean trackCameraPosition = false;
    private boolean myLocationEnabled = false;
    private boolean myLocationButtonEnabled = false;
    private boolean indoorEnabled = true;
    private boolean trafficEnabled = false;
    private boolean disposed = false;
    private final float density;
    private MethodChannel.Result mapReadyResult;
    private final int registrarActivityHashCode;
    private final Context context;
    private final MarkersController markersController;
    private final PolygonsController polygonsController;
    private final PolylinesController polylinesController;
    private final CirclesController circlesController;
    private List<Object> initialMarkers;
    private List<Object> initialPolygons;
    private List<Object> initialPolylines;
    private List<Object> initialCircles;


    AmapMapController(
            int id,
            Context context,
            AtomicInteger activityState,
            PluginRegistry.Registrar registrar,
            AMapOptions options) {

        this.id = id;
        this.context = context;
        this.activityState = activityState;
        this.registrar = registrar;
        this.mapView = new MapView(context, options);
        this.density = context.getResources().getDisplayMetrics().density;
        methodChannel =
                new MethodChannel(registrar.messenger(), "amap.mobileconnected.com/map_" + id);
        methodChannel.setMethodCallHandler(this);
        this.registrarActivityHashCode = registrar.activity().hashCode();
        this.markersController = new MarkersController(methodChannel);
        this.polygonsController = new PolygonsController(methodChannel);
        this.polylinesController = new PolylinesController(methodChannel, density);
        this.circlesController = new CirclesController(methodChannel);
    }

    void init() {
        switch (activityState.get()) {
            case STOPPED:
            case PAUSED:
                mapView.onCreate(null);
                mapView.onResume();
                mapView.onPause();
                break;
            case RESUMED:
                mapView.onCreate(null);
                mapView.onResume();
                break;
            case STARTED:
                mapView.onCreate(null);
                break;
            case CREATED:
                mapView.onCreate(null);
                break;
            case DESTROYED:
                // Nothing to do, the activity has been completely destroyed.
                break;
            default:
                throw new IllegalArgumentException(
                        "Cannot interpret " + activityState.get() + " as an activity state");
        }
        registrar.activity().getApplication().registerActivityLifecycleCallbacks(this);

        amapMap = mapView.getMap();
        amapMap.showIndoorMap(this.indoorEnabled);
        amapMap.setTrafficEnabled(trafficEnabled);
        amapMap.setOnInfoWindowClickListener(this);
        //TODO 评估 amapMap.setOnCameraMoveStartedListener(this)
        //        amapMap.setOnCameraMoveListener(this)
        //        amapMap.setOnCameraIdleListener(this)被替换的影响
        amapMap.setOnCameraChangeListener(this);
        amapMap.setOnMarkerClickListener(this);
        amapMap.setOnMarkerDragListener(this);
        //TODO rm since Amap not supported
//        amapMap.setOnPolygonClickListener(this)
        amapMap.setOnPolylineClickListener(this);
        //TODO rm since Amap not supported
//        amapMap.setOnCircleClickListener(this)
        amapMap.setOnMapClickListener(this);
        amapMap.setOnMapLongClickListener(this);
        updateMyLocationSettings();
        markersController.setAmapMap(amapMap);
        polygonsController.setAmapMap(amapMap);
        polylinesController.setAmapMap(amapMap);
        circlesController.setAmapMap(amapMap);

        updateInitialMarkers();
        updateInitialPolygons();
        updateInitialPolylines();
        updateInitialCircles();
    }

    @Override
    public void onActivityCreated(Activity activity, Bundle savedInstanceState) {
        if (!disposed && activity.hashCode() == registrarActivityHashCode) {
            mapView.onCreate(savedInstanceState);
        }
    }

    @Override
    public void onActivityStarted(Activity activity) {}

    @Override
    public void onActivityResumed(Activity activity) {
        if (!disposed && activity.hashCode() == registrarActivityHashCode) {
            mapView.onResume();
        }
    }

    @Override
    public void onActivityPaused(Activity activity) {
        if (!disposed && activity.hashCode() == registrarActivityHashCode) {
            mapView.onPause();
        }
    }

    @Override
    public void onActivityStopped(Activity activity) {
    }

    @Override
    public void onActivitySaveInstanceState(Activity activity, Bundle outState) {
        if (!disposed && activity.hashCode() == registrarActivityHashCode) {
            mapView.onSaveInstanceState(outState);
        }
    }

    @Override
    public void onActivityDestroyed(Activity activity) {
        if (!disposed && activity.hashCode() == registrarActivityHashCode) {
            mapView.onDestroy();
        }
    }

    //TODO rm since Amap not support,but the impact needs to be assessed
//    override fun onCameraMoveStarted(reason: Int) {
//        val arguments = HashMap<String, Any>(2)
//        val isGesture = reason == AmapMap.OnCameraMoveStartedListener.REASON_GESTURE
//        arguments["isGesture"] = isGesture
//        methodChannel.invokeMethod("camera#onMoveStarted", arguments)
//    }

    @Override
    public void onCameraChange(CameraPosition cameraPosition) {
        if (!trackCameraPosition) {
            return;
        }
        final Map<String, Object> arguments = new HashMap<>(2);
        arguments.put("position", Convert.cameraPositionToJson(amapMap.getCameraPosition()));
        methodChannel.invokeMethod("camera#onMove", arguments);
    }

    @Override
    public void onCameraChangeFinish(CameraPosition cameraPosition) {
        methodChannel.invokeMethod("camera#onIdle", Collections.singletonMap("map", id));
    }

    @Override
    public void onInfoWindowClick(Marker marker) {
        markersController.onInfoWindowTap(marker.getId());
    }

    @Override
    public void onMapClick(LatLng latLng) {
        final Map<String, Object> arguments = new HashMap<>(2);
        arguments.put("position", Convert.latLngToJson(latLng));
        methodChannel.invokeMethod("map#onTap", arguments);
    }

    @Override
    public void onMapLongClick(LatLng latLng) {
        final Map<String, Object> arguments = new HashMap<>(2);
        arguments.put("position", Convert.latLngToJson(latLng));
        methodChannel.invokeMethod("map#onLongPress", arguments);
    }

    @Override
    public boolean onMarkerClick(Marker marker) {
        return markersController.onMarkerTap(marker.getId());
    }

    @Override
    public void onMarkerDragStart(Marker marker) {
    }

    @Override
    public void onMarkerDrag(Marker marker) {
    }

    @Override
    public void onMarkerDragEnd(Marker marker) {
        markersController.onMarkerDragEnd(marker.getId(), marker.getPosition());
    }

    @Override
    public void onPolylineClick(Polyline polyline) {
        polylinesController.onPolylineTap(polyline.getId());
    }

    @Override
    public void setCameraTargetBounds(LatLngBounds bounds) {
        //相对于原来的操作，这里会执行移动相机的操作
        amapMap.setMapStatusLimits(bounds);
    }

    @Override
    public void setCompassEnabled(boolean compassEnabled) {
        amapMap.getUiSettings().setCompassEnabled(compassEnabled);
    }

    @Override
    public void setMapToolbarEnabled(boolean setMapToolbarEnabled) {
        //TODO Amap not supported
    }

    @Override
    public void setMapType(int mapType) {
        //TODO check whether mapType of google used for Amap
        amapMap.setMapType(mapType);
    }

    @Override
    public void setMinMaxZoomPreference(Float min, Float max) {
        amapMap.resetMinMaxZoomPreference();
        if (min != null) {
            amapMap.setMinZoomLevel(min);
        }
        if (max != null) {
            amapMap.setMaxZoomLevel(max);
        }
    }

    @Override
    public void setPadding(float top, float left, float bottom, float right) {
        //TODO padding & CameraTargetBounds must be done together with move camera
        //so this need to changed
    }

    @Override
    public void setRotateGesturesEnabled(boolean rotateGesturesEnabled) {
        amapMap.getUiSettings().setRotateGesturesEnabled(rotateGesturesEnabled);
    }

    @Override
    public void setScrollGesturesEnabled(boolean scrollGesturesEnabled) {
        amapMap.getUiSettings().setScrollGesturesEnabled(scrollGesturesEnabled);
    }

    @Override
    public void setTiltGesturesEnabled(boolean tiltGesturesEnabled) {
        amapMap.getUiSettings().setTiltGesturesEnabled(tiltGesturesEnabled);
    }

    @Override
    public void setTrackCameraPosition(boolean trackCameraPosition) {
        this.trackCameraPosition = trackCameraPosition;
    }

    @Override
    public void setZoomGesturesEnabled(boolean zoomGesturesEnabled) {
        amapMap.getUiSettings().setZoomGesturesEnabled(zoomGesturesEnabled);
    }

    @Override
    public void setMyLocationEnabled(boolean myLocationEnabled) {
        if (this.myLocationEnabled == myLocationEnabled) {
            return;
        }
        this.myLocationEnabled = myLocationEnabled;
        updateMyLocationSettings();
    }

    @Override
    public void setMyLocationButtonEnabled(boolean myLocationButtonEnabled) {
        if (this.myLocationButtonEnabled == myLocationButtonEnabled) {
            return;
        }
        this.myLocationButtonEnabled = myLocationButtonEnabled;
        updateMyLocationSettings();
    }

    @Override
    public void setIndoorEnabled(boolean indoorEnabled) {
        this.indoorEnabled = indoorEnabled;
    }

    @Override
    public void setTrafficEnabled(boolean trafficEnabled) {
        this.trafficEnabled = trafficEnabled;
    }

    @Override
    public void setInitialMarkers(Object initialMarkers) {
        this.initialMarkers = (List<Object>) initialMarkers;
        if (amapMap != null) {
            updateInitialMarkers();
        }
    }

    private void updateInitialMarkers() {
        markersController.addMarkers(initialMarkers);
    }

    @Override
    public void setInitialPolygons(Object initialPolygons) {
        this.initialPolygons = (List<Object>) initialPolygons;
        if (amapMap != null) {
            updateInitialPolygons();
        }
    }

    private void updateInitialPolygons() {
        polygonsController.addPolygons(initialPolygons);
    }

    @Override
    public void setInitialPolylines(Object initialPolylines) {
        this.initialPolylines = (List<Object>) initialPolylines;
        if (amapMap != null) {
            updateInitialPolylines();
        }
    }

    private void updateInitialPolylines() {
        polylinesController.addPolylines(initialPolylines);
    }

    @Override
    public void setInitialCircles(Object initialCircles) {
        this.initialCircles = (List<Object>) initialCircles;
        if (amapMap != null) {
            updateInitialCircles();
        }
    }

    private void updateInitialCircles() {
        circlesController.addCircles(initialCircles);
    }

    private CameraPosition getCameraPosition() {
        return trackCameraPosition ? amapMap.getCameraPosition() : null;
    }

    private void moveCamera(CameraUpdate cameraUpdate) {
        amapMap.moveCamera(cameraUpdate);
    }

    private void animateCamera(CameraUpdate cameraUpdate) {
        amapMap.animateCamera(cameraUpdate);
    }

    @Override
    public void onMethodCall(MethodCall methodCall, MethodChannel.Result result) {
        switch (methodCall.method) {
//            TODO 这个和地图的初始化有关，但Amap不需要这个回调，所以有待看着这个但影响，再看是否删除
            case "map#waitForMap":
                if (amapMap != null) {
                    result.success(null);
                    return;
                }
                mapReadyResult = result;
                break;
            case "map#update":
            {
                Convert.interpretGoogleMapOptions(methodCall.argument("options"), this);
                result.success(Convert.cameraPositionToJson(getCameraPosition()));
                break;
            }
            case "map#getVisibleRegion":
            {
                if (amapMap != null) {
                    LatLngBounds latLngBounds = amapMap.getProjection().getVisibleRegion().latLngBounds;
                    result.success(Convert.latlngBoundsToJson(latLngBounds));
                } else {
                    result.error(
                            "GoogleMap uninitialized",
                            "getVisibleRegion methodCalled prior to map initialization",
                            null);
                }
                break;
            }
            case "map#getScreenCoordinate":
            {
                if (amapMap != null) {
                    LatLng latLng = Convert.toLatLng(methodCall.arguments);
                    Point screenLocation = amapMap.getProjection().toScreenLocation(latLng);
                    result.success(Convert.pointToJson(screenLocation));
                } else {
                    result.error(
                            "GoogleMap uninitialized",
                            "getScreenCoordinate methodCalled prior to map initialization",
                            null);
                }
                break;
            }
            case "map#getLatLng":
            {
                if (amapMap != null) {
                    Point point = Convert.toPoint(methodCall.arguments);
                    LatLng latLng = amapMap.getProjection().fromScreenLocation(point);
                    result.success(Convert.latLngToJson(latLng));
                } else {
                    result.error(
                            "GoogleMap uninitialized", "getLatLng methodCalled prior to map initialization", null);
                }
                break;
            }
            case "camera#move":
            {
                final CameraUpdate cameraUpdate =
                        Convert.toCameraUpdate(methodCall.argument("cameraUpdate"), density);
                moveCamera(cameraUpdate);
                result.success(null);
                break;
            }
            case "camera#animate":
            {
                final CameraUpdate cameraUpdate =
                        Convert.toCameraUpdate(methodCall.argument("cameraUpdate"), density);
                animateCamera(cameraUpdate);
                result.success(null);
                break;
            }
            case "markers#update":
            {
                Object markersToAdd = methodCall.argument("markersToAdd");
                markersController.addMarkers((List<Object>) markersToAdd);
                Object markersToChange = methodCall.argument("markersToChange");
                markersController.changeMarkers((List<Object>) markersToChange);
                Object markerIdsToRemove = methodCall.argument("markerIdsToRemove");
                markersController.removeMarkers((List<Object>) markerIdsToRemove);
                result.success(null);
                break;
            }
            case "polygons#update":
            {
                Object polygonsToAdd = methodCall.argument("polygonsToAdd");
                polygonsController.addPolygons((List<Object>) polygonsToAdd);
                Object polygonsToChange = methodCall.argument("polygonsToChange");
                polygonsController.changePolygons((List<Object>) polygonsToChange);
                Object polygonIdsToRemove = methodCall.argument("polygonIdsToRemove");
                polygonsController.removePolygons((List<Object>) polygonIdsToRemove);
                result.success(null);
                break;
            }
            case "polylines#update":
            {
                Object polylinesToAdd = methodCall.argument("polylinesToAdd");
                polylinesController.addPolylines((List<Object>) polylinesToAdd);
                Object polylinesToChange = methodCall.argument("polylinesToChange");
                polylinesController.changePolylines((List<Object>) polylinesToChange);
                Object polylineIdsToRemove = methodCall.argument("polylineIdsToRemove");
                polylinesController.removePolylines((List<Object>) polylineIdsToRemove);
                result.success(null);
                break;
            }
            case "circles#update":
            {
                Object circlesToAdd = methodCall.argument("circlesToAdd");
                circlesController.addCircles((List<Object>) circlesToAdd);
                Object circlesToChange = methodCall.argument("circlesToChange");
                circlesController.changeCircles((List<Object>) circlesToChange);
                Object circleIdsToRemove = methodCall.argument("circleIdsToRemove");
                circlesController.removeCircles((List<Object>) circleIdsToRemove);
                result.success(null);
                break;
            }
            case "map#isCompassEnabled":
            {
                result.success(amapMap.getUiSettings().isCompassEnabled());
                break;
            }
            //TODO 这个在Amap里面找不到对应的
            case "map#isMapToolbarEnabled":
            {
//                result.success(amapMap.getUiSettings().isMapToolbarEnabled());
                result.notImplemented();
                break;
            }
            case "map#getMinMaxZoomLevels":
            {
                List<Float> zoomLevels = new ArrayList<>(2);
                zoomLevels.add(amapMap.getMinZoomLevel());
                zoomLevels.add(amapMap.getMaxZoomLevel());
                result.success(zoomLevels);
                break;
            }
            case "map#isZoomGesturesEnabled":
            {
                result.success(amapMap.getUiSettings().isZoomGesturesEnabled());
                break;
            }
            case "map#isScrollGesturesEnabled":
            {
                result.success(amapMap.getUiSettings().isScrollGesturesEnabled());
                break;
            }
            case "map#isTiltGesturesEnabled":
            {
                result.success(amapMap.getUiSettings().isTiltGesturesEnabled());
                break;
            }
            case "map#isRotateGesturesEnabled":
            {
                result.success(amapMap.getUiSettings().isRotateGesturesEnabled());
                break;
            }
            case "map#isMyLocationButtonEnabled":
            {
                result.success(amapMap.getUiSettings().isMyLocationButtonEnabled());
                break;
            }
            case "map#isTrafficEnabled":
            {
                result.success(amapMap.isTrafficEnabled());
                break;
            }
            case "map#setStyle":
            {
                String mapStyle = (String) methodCall.arguments;
                boolean mapStyleSet;
                //TODO Amap必须使用setCustomMapStyle(CustomMapStyleOptions())
                //TODO 所以这里的mapStyle必须是高得地图配置的线上效果，https://lbs.amap.com/api/android-sdk/guide/create-map/custom
                try {
                    amapMap.setCustomMapStyle(new CustomMapStyleOptions().setStyleId(mapStyle));
                    mapStyleSet = true;
                } catch (Exception e) {
                    e.printStackTrace();
                    mapStyleSet = false;
                }
                ArrayList<Object> mapStyleResult = new ArrayList<>(2);
                mapStyleResult.add(mapStyleSet);
                if (!mapStyleSet) {
                    mapStyleResult.add(
                            "Unable to set the map style. Please check console logs for errors.");
                }
                result.success(mapStyleResult);
                break;
            }
            default:
                result.notImplemented();
        }
    }

    @Override
    public View getView() {
        return mapView;
    }

    @Override
    public void dispose() {
        if (disposed) {
            return;
        }
        disposed = true;
        methodChannel.setMethodCallHandler(null);
        mapView.onDestroy();
        registrar.activity().getApplication().unregisterActivityLifecycleCallbacks(this);
    }

    @SuppressLint("MissingPermission")
    private void updateMyLocationSettings() {
        if (hasLocationPermission()) {
            // The plugin doesn't add the location permission by default so that apps that don't need
            // the feature won't require the permission.
            // Gradle is doing a static check for missing permission and in some configurations will
            // fail the build if the permission is missing. The following disables the Gradle lint.
            //noinspection ResourceType
            amapMap.setMyLocationEnabled(myLocationEnabled);
            amapMap.getUiSettings().setMyLocationButtonEnabled(myLocationButtonEnabled);
        } else {
            // TODO(amirh): Make the options update fail.
            // https://github.com/flutter/flutter/issues/24327
            Log.e(TAG, "Cannot enable MyLocation layer as location permissions are not granted");
        }
    }

    private boolean hasLocationPermission() {
        return checkSelfPermission(Manifest.permission.ACCESS_FINE_LOCATION)
                == PackageManager.PERMISSION_GRANTED
                || checkSelfPermission(Manifest.permission.ACCESS_COARSE_LOCATION)
                == PackageManager.PERMISSION_GRANTED;
    }

    private int checkSelfPermission(String permission) {
        if (permission == null) {
            throw new IllegalArgumentException("permission is null");
        }
        return context.checkPermission(
                permission, android.os.Process.myPid(), android.os.Process.myUid());
    }
}
