//
//  NBMapCircleController.m
//  nb_map
//
//  Created by Guo Longxiang, (Longxiang.Guo@partner.bmw.com) on 2019/10/29.
//

#import "NBMapCircleController.h"
#import "NBJsonConversions.h"
#import "MAMapView+CenterAroundVisiableOverlays.h"

@implementation NBMapCircleController{
    MAMapView* _mapView;
    NBCircle* _circle;
    MACircleRenderer* _circleRenderer;
}

-(instancetype)initCircleWithPosition:(CLLocationCoordinate2D)position radius:(CLLocationDistance)radius circleId:(NSString *)circleId mapView:(MAMapView *)mapView {
    self = [super init];
    if(self){
        _circle = [NBCircle circleWithCenterCoordinate:position radius:radius];
        _circle.accessibilityLabel = circleId;
        _circle.tappable = false;
        _circle.visible = true;
        _circle.fillColor = kMAOverlayRendererDefaultFillColor;
        _circle.strokeColor = kMAOverlayRendererDefaultStrokeColor;
        _circle.level = MAOverlayLevelAboveLabels;
        _circle.lineWidth = 0;
        _circleId = circleId;
        _mapView = mapView;
    }
    return self;
}

#pragma mark - Help Methods
- (MACircleRenderer*) circleRender {
    if (_circleRenderer == nil) {
       _circleRenderer = (MACircleRenderer*)[_mapView rendererForOverlay:_circle];
    }
    return _circleRenderer;
}

-(void)removeCircle {
    [_mapView removeOverlay:_circle];
    _circleRenderer = nil;
    [_mapView updateCachedOverlay:_circle event:NBVisiableOverlayDelete];
}

- (void)addCircle {
    [_mapView addOverlay:_circle level:_circle.level];
    if (_circle.isVisiable) {
        [_mapView updateCachedOverlay:_circle event:NBVisiableOverlayAdd];
    }
}

#pragma mark - NBMapCircleOptionsSink methods
- (void)setConsumeTapEvents:(BOOL)consumes {
    _circle.tappable = consumes;
}
//Notice: If set the overlay to hidden, we should hidden the view and ajusted the overlay postion in mapview.
- (void)setVisible:(BOOL)visible {
    if (_circle.isVisiable){
        [_mapView updateCachedOverlay:_circle event:NBVisiableOverlayAdd];
    } else{
        [_mapView updateCachedOverlay:_circle event:NBVisiableOverlayDelete];
    }
    if (_circle.visible == visible) {
        return;
    }
    if (![self circleRender]) {
        return;
    }
    CGFloat alpha = visible ? 1.0 : 0.0;
    UIColor* newFillColor = [[_circleRenderer.fillColor copy]colorWithAlphaComponent:alpha];
    UIColor* newStrokeColor = [[_circleRenderer.strokeColor copy]colorWithAlphaComponent:alpha];
    _circleRenderer.fillColor = newFillColor;
    _circleRenderer.strokeColor = newStrokeColor;
    _circle.fillColor = newFillColor;
    _circle.strokeColor = newStrokeColor;
    _circle.visible = visible;
}

- (void)setCenter:(CLLocationCoordinate2D)center {
    [_circle setCoordinate:center];
}

- (void)setRadius:(CLLocationDistance)radius {
    _circle.radius = radius;
}

- (void)setStrokeColor:(UIColor*)color {
    _circle.strokeColor = color;
    if (![self circleRender]) {
        return;
    }
    [self circleRender].strokeColor = color;
}

- (void)setStrokeWidth:(CGFloat)width {
    _circle.lineWidth = width;
    if (![self circleRender]) {
        return;
    }
    [self circleRender].lineWidth = width;
}

- (void)setFillColor:(UIColor*)color {
    _circle.fillColor = color;
    if (![self circleRender]) {
        return;
    }
    [self circleRender].fillColor = color;
}

- (void)setNeedsUpdate {
    if (![self circleRender]) {
        return;
    }
    [[self circleRender] setNeedsUpdate];
}

- (void)setZIndex:(int)zIndex {
    if (zIndex >= 1) {
        _circle.level = MAOverlayLevelAboveLabels;
    } else {
        _circle.level = MAOverlayLevelAboveRoads;
    }
}


@end

static int ToInt(NSNumber* data) {
    return [NBJsonConversions toInt:data];
}
static BOOL ToBool(NSNumber* data) {
    return [NBJsonConversions toBool:data];
}
static CLLocationCoordinate2D ToLocation(NSArray* data) {
    return [NBJsonConversions toLocation:data];
}
static CLLocationDistance ToDistance(NSNumber* data) {
    return [NBJsonConversions toFloat:data];
}
static UIColor* ToColor(NSNumber* data){
    return [NBJsonConversions toColor:data];
}


static void InterpretCircleOptions(NSDictionary* data,
                                   id<NBMapCircleOptionsSink> sink,
                                   NSObject<FlutterPluginRegistrar>* registrar) {
    NSNumber* consumeTapEvents = data[@"consumeTapEvents"];
    if (consumeTapEvents) {
        [sink setConsumeTapEvents:ToBool(consumeTapEvents)];
    }
    NSNumber* visible = data[@"visible"];
    if (visible) {
        [sink setVisible:ToBool(visible)];
    }
    NSArray* center = data[@"center"];
    if (center) {
        [sink setCenter:ToLocation(center)];
    }
    
    NSNumber* radius = data[@"radius"];
    if (radius) {
        [sink setRadius:ToDistance(radius)];
    }
    
    NSNumber* strokeColor = data[@"strokeColor"];
    if (strokeColor) {
        [sink setStrokeColor:ToColor(strokeColor)];
    }
    
    NSNumber* strokeWidth = data[@"strokeWidth"];
    if (strokeWidth) {
        [sink setStrokeWidth:ToInt(strokeWidth)];
    }
    
    NSNumber* fillColor = data[@"fillColor"];
    if (fillColor) {
        [sink setFillColor:ToColor(fillColor)];
    }
    [sink setNeedsUpdate];
}

@implementation NBMapCirclesController {
    NSMutableDictionary* _circleIdToController;
    FlutterMethodChannel* _methodChannel;
    NSObject<FlutterPluginRegistrar>* _registrar;
    MAMapView* _mapView;
}

-(instancetype)init:(FlutterMethodChannel *)methodChannel mapView:(MAMapView *)mapView registrar:(NSObject<FlutterPluginRegistrar> *)registrar {
    self = [super init];
    if (self){
        _methodChannel = methodChannel;
        _registrar = registrar;
        _mapView = mapView;
        _circleIdToController = [NSMutableDictionary dictionaryWithCapacity:1];
    }
    return self;
}

#pragma mark - Circle Update Events
- (void)addCircles:(NSArray *)circlesToAdd {
    for (NSDictionary* circle in circlesToAdd) {
        CLLocationCoordinate2D position = [NBMapCirclesController getPosition:circle];
        CLLocationDistance radius = [NBMapCirclesController getRadius:circle];
        NSString* circleId = [NBMapCirclesController getCircleId:circle];
        NBMapCircleController* controller = [[NBMapCircleController alloc]initCircleWithPosition:position radius:radius circleId:circleId mapView:_mapView];
        InterpretCircleOptions(circle, controller, _registrar);
        [controller addCircle];
        _circleIdToController[circleId] = controller;
    }
    ///TODO:批量添加 VS 逐个添加
 
}

- (void)changeCircles:(NSArray *)circlesToChange {
    for (NSDictionary* circle in circlesToChange) {
        NSString* circleId = [NBMapCirclesController getCircleId:circle];
        NBMapCircleController* controller = _circleIdToController[circleId];
        if (!controller) {
            continue;
        }
        InterpretCircleOptions(circle, controller, _registrar);
    }
}

- (void)removeCircleIds:(NSArray *)circleIdsToRemove {
    for (NSString* circleId in circleIdsToRemove) {
        if (!circleId) {
            continue;
        }
        NBMapCircleController* controller = _circleIdToController[circleId];
        if (!controller) {
            continue;
        }
        [controller removeCircle];
        [_circleIdToController removeObjectForKey:circleId];
    }
}

- (bool)hasCircleWithId:(NSString *)circleId {
    if (!circleId) {
        return false;
    }
    return _circleIdToController[circleId] != nil;
}

- (void)onCircleTap:(NSString *)circleId {
    if (!circleId) {
        return;
    }
    NBMapCircleController* controller = _circleIdToController[circleId];
    if (!controller) {
        return;
    }
    [_methodChannel invokeMethod:@"circle#onTap" arguments:@{@"circleId":circleId}];
}

#pragma mark - Help Methods
+ (CLLocationCoordinate2D)getPosition:(NSDictionary*)circle {
    NSArray* center = circle[@"center"];
    return ToLocation(center);
}

+ (CLLocationDistance)getRadius:(NSDictionary*)circle {
    NSNumber* radius = circle[@"radius"];
    return ToDistance(radius);
}

+ (NSString*)getCircleId:(NSDictionary*)circle {
    return circle[@"circleId"];
}

@end

@implementation NBCircle

@end
