//
//  NBMapController.m
//  nb_map
//
//  Created by Guo Longxiang, (Longxiang.Guo@partner.bmw.com) on 2019/10/29.
//

#import "NBMapController.h"
#import "NBMapCircleController.h"
#import "NBMapMarkerController.h"
#import "NBMapPolylineController.h"
#import "NBMapPolygonController.h"
#import "NBJsonConversions.h"
#import "NBCommonUtility.h"
#import "NBRendersFactory.h"
#import "MAMapView+CenterAroundVisiableOverlays.h"
#import "NBMarkerViewsFactory.h"
/**
 兼容Google API的类库对应关系
 GMSCameraPosition: MAMapStatus  MACoordinateBounds
 GMSCoordinateBounds: MACoordinateBounds
 GMSCameraUpdate: MACameraUpdate
 */
#ifndef PIN_SIZE
#define PIN_SIZE 45.0
#endif
#ifndef PADDING
#define PADDING (PIN_SIZE*0.5 + 40.0) //Notice: padding 和 marker的锚点对应,需根据实际情况添加
#endif
#ifndef ZOOM_RATIO
#define ZOOM_RATIO 0.1
#endif
static NSDictionary* MapStatusToJson(MAMapStatus* mapStatus);
static NSDictionary* PointToJson(CGPoint point);
static NSArray* LocationToJson(CLLocationCoordinate2D position);
static CLLocationCoordinate2D ToLocation(NSArray* data);
static CGPoint ToCGPoint(NSDictionary* data);
static CGPoint ToPivot(NSArray* data);
static float ToFloat(NSNumber* data);
static double ToDouble(NSNumber* data);
static MAMapStatus* ToMapStatus(NSDictionary* data);
static MAMapStatus* ToOptionalMapStatus(NSDictionary* data);
static MACoordinateBounds ToBounds(NSArray* data);
static MACoordinateBounds ToOptionalBounds(NSArray* data);
static MACoordinateBounds RegionToBounds(MACoordinateRegion region);
static MACoordinateRegion BoundsToRegion(MACoordinateBounds bounds);
static NSDictionary* MACoordinateBoundsToJson(MACoordinateBounds bounds);
static void InterpretMapOptions(NSDictionary* data, id<NBMapOptionsSink> sink);

@implementation NBMapFactory{
    NSObject<FlutterPluginRegistrar>* _registrar;
}

- (instancetype)initWithRegistrar:(NSObject<FlutterPluginRegistrar>*)registrar {
    self = [super init];
    if (self) {
        
        _registrar = registrar;
    }
    return self;
}

- (NSObject<FlutterMessageCodec>*)createArgsCodec {
    return [FlutterStandardMessageCodec sharedInstance];
}

- (NSObject<FlutterPlatformView>*)createWithFrame:(CGRect)frame
                                   viewIdentifier:(int64_t)viewId
                                        arguments:(id _Nullable)args {
    return [[NBMapController alloc] initWithFrame:frame
                                   viewIdentifier:viewId
                                        arguments:args
                                        registrar:_registrar];
}
@end

@implementation NBMapController {
    MAMapView* _mapView;
    int64_t _viewId;
    FlutterMethodChannel* _channel;
    BOOL _trackCameraPosition;
    NSObject<FlutterPluginRegistrar>* _registrar;
    // Used for the temporary workaround for a bug that the camera is not properly positioned at
    // initialization. https://github.com/flutter/flutter/issues/24806
    // TODO(cyanglaz): Remove this temporary fix once the Maps SDK issue is resolved.
    // https://github.com/flutter/flutter/issues/27550
    BOOL _cameraDidInitialSetup;
    NBMapMarkersController* _markersController;
    NBMapPolygonsController* _polygonsController;
    NBMapPolylinesController* _polylinesController;
    NBMapCirclesController* _circlesController;
}

- (instancetype)initWithFrame:(CGRect)frame
               viewIdentifier:(int64_t)viewId
                    arguments:(id _Nullable)args
                    registrar:(NSObject<FlutterPluginRegistrar>*)registrar {
    if ([super init]) {
        _viewId = viewId;
        NSLog(@"===== MapView Create ViewId %lld args: \n %@", viewId, args);
        MAMapStatus* mapStatus = ToOptionalMapStatus(args[@"initialCameraPosition"]);
        _mapView = [[MAMapView alloc]initWithFrame:frame];
        _mapView.accessibilityElementsHidden = NO;
        _trackCameraPosition = NO;
        [_mapView setMapStatus:mapStatus animated:false];
        InterpretMapOptions(args[@"options"], self);
        NSString* channelName =
        [NSString stringWithFormat:@"plugins.flutter.io/nb_maps_%lld", viewId];
        _channel = [FlutterMethodChannel methodChannelWithName:channelName
                                               binaryMessenger:registrar.messenger];
        __weak __typeof__(self) weakSelf = self;
        [_channel setMethodCallHandler:^(FlutterMethodCall* call, FlutterResult result) {
            if (weakSelf) {
                [weakSelf onMethodCall:call result:result];
            }
        }];
        _mapView.delegate = self;
        _registrar = registrar;
        _cameraDidInitialSetup = NO;
        _markersController = [[NBMapMarkersController alloc] init:_channel
                                                          mapView:_mapView
                                                        registrar:registrar];
        _polygonsController = [[NBMapPolygonsController alloc] init:_channel
                                                            mapView:_mapView
                                                          registrar:registrar];
        _polylinesController = [[NBMapPolylinesController alloc] init:_channel
                                                              mapView:_mapView
                                                            registrar:registrar];
        _circlesController = [[NBMapCirclesController alloc] init:_channel
                                                          mapView:_mapView
                                                        registrar:registrar];
        id markersToAdd = args[@"markersToAdd"];
        if ([markersToAdd isKindOfClass:[NSArray class]]) {
            [_markersController addMarkers:markersToAdd];
        }
        id polygonsToAdd = args[@"polygonsToAdd"];
        if ([polygonsToAdd isKindOfClass:[NSArray class]]) {
            [_polygonsController addPolygons:polygonsToAdd];
        }
        id polylinesToAdd = args[@"polylinesToAdd"];
        if ([polylinesToAdd isKindOfClass:[NSArray class]]) {
            [_polylinesController addPolylines:polylinesToAdd];
        }
        id circlesToAdd = args[@"circlesToAdd"];
        if ([circlesToAdd isKindOfClass:[NSArray class]]) {
            [_circlesController addCircles:circlesToAdd];
        }
        [_mapView centerWithRule:NBCenterMapAroundAll
                         padding:UIEdgeInsetsMake(PADDING, PADDING + 30.0, PADDING - 30.0, PADDING)
                        animated:false];
        CGFloat x = _mapView.bounds.size.width;
        CGFloat y = _mapView.bounds.size.width;
        NBUserLocationButton* button = [[NBUserLocationButton alloc]initWithFrame:CGRectMake( x - 60.0, y-60.0, 40.0, 40.0)];
        button.hidden = true;
        [_mapView addSubview:button];
        __block typeof(self) strongSelf = self;
        button.center = ^{
            [strongSelf->_mapView setCenterCoordinate: self->_mapView.userLocation.coordinate];
        };
    }
    return self;
}

- (UIView*)view {
    return _mapView;
}

- (void)onMethodCall:(FlutterMethodCall*)call result:(FlutterResult)result {
    NSLog(@"====== OnMethod ======== \n %@  args: \n%@", call.method, call.arguments);
    if ([call.method isEqualToString:@"map#show"]) {
        [self showAtX:ToDouble(call.arguments[@"x"]) Y:ToDouble(call.arguments[@"y"])];
        result(nil);
    } else if ([call.method isEqualToString:@"map#hide"]) {
        [self hide];
        result(nil);
    } else if ([call.method isEqualToString:@"camera#animate"]) {
        [self updateCamera:call.arguments[@"cameraUpdate"] animate:true];
        result(nil);
    } else if ([call.method isEqualToString:@"camera#move"]) {
        [self updateCamera:call.arguments[@"cameraUpdate"] animate:false];
        result(nil);
    } else if ([call.method isEqualToString:@"map#update"]) {
        InterpretMapOptions(call.arguments[@"options"], self);
        result(MapStatusToJson([self cameraPosition]));
    } else if ([call.method isEqualToString:@"map#getVisibleRegion"]) {
        if (_mapView != nil) {
            MACoordinateRegion visiableRegion = _mapView.region;
            MACoordinateBounds bounds = RegionToBounds(visiableRegion);
            result(MACoordinateBoundsToJson(bounds));
        } else {
            result([FlutterError errorWithCode:@"NbMap uninitialized"
                                       message:@"getVisibleRegion called prior to map initialization"
                                       details:nil]);
        }
    } else if ([call.method isEqualToString:@"map#getScreenCoordinate"]) {
        if (_mapView != nil && call.arguments) {
            CLLocationCoordinate2D coordinate = [NBJsonConversions toLocation:call.arguments];
            CGPoint point = [_mapView convertCoordinate:coordinate toPointToView:_mapView];
            result(PointToJson(point));
        } else {
            result([FlutterError errorWithCode:@"GoogleMap uninitialized"
                                       message:@"getScreenCoordinate called prior to map initialization"
                                       details:nil]);
        }
    } else if ([call.method isEqualToString:@"map#getLatLng"]) {
        if (_mapView != nil && call.arguments) {
            CGPoint point = ToCGPoint(call.arguments);
            CLLocationCoordinate2D latlng = [_mapView convertPoint:point toCoordinateFromView:_mapView];
            result(LocationToJson(latlng));
        } else {
            result([FlutterError errorWithCode:@"GoogleMap uninitialized"
                                       message:@"getLatLng called prior to map initialization"
                                       details:nil]);
        }
    } else if ([call.method isEqualToString:@"map#waitForMap"]) {
        result(nil);
    } else if ([call.method isEqualToString:@"markers#update"]) {
        id markersToAdd = call.arguments[@"markersToAdd"];
        if ([markersToAdd isKindOfClass:[NSArray class]]) {
            [_markersController addMarkers:markersToAdd];
        }
        id markersToChange = call.arguments[@"markersToChange"];
        if ([markersToChange isKindOfClass:[NSArray class]]) {
            [_markersController changeMarkers:markersToChange];
        }
        id markerIdsToRemove = call.arguments[@"markerIdsToRemove"];
        if ([markerIdsToRemove isKindOfClass:[NSArray class]]) {
            [_markersController removeMarkerIds:markerIdsToRemove];
        }
        [_mapView centerWithRule:NBCenterMapAroundAll
                         padding:UIEdgeInsetsMake(PADDING, PADDING, PADDING, PADDING)
                        animated:true];
        result(nil);
        //TODO: Fix me center the map around pins
        NSLog(@"mapView overlays count: %lu", (unsigned long)_mapView.overlays.count);
        [_mapView showOverlays:_mapView.overlays animated:false];
    } else if ([call.method isEqualToString:@"polygons#update"]) {
        id polygonsToAdd = call.arguments[@"polygonsToAdd"];
        if ([polygonsToAdd isKindOfClass:[NSArray class]]) {
            [_polygonsController addPolygons:polygonsToAdd];
        }
        id polygonsToChange = call.arguments[@"polygonsToChange"];
        if ([polygonsToChange isKindOfClass:[NSArray class]]) {
            [_polygonsController changePolygons:polygonsToChange];
        }
        id polygonIdsToRemove = call.arguments[@"polygonIdsToRemove"];
        if ([polygonIdsToRemove isKindOfClass:[NSArray class]]) {
            [_polygonsController removePolygonIds:polygonIdsToRemove];
        }
        [_mapView centerWithRule:NBCenterMapAroundAll
                         padding:UIEdgeInsetsMake(PADDING, PADDING, PADDING, PADDING)
                        animated:true];
        result(nil);
    } else if ([call.method isEqualToString:@"polylines#update"]) {
        id polylinesToAdd = call.arguments[@"polylinesToAdd"];
        if ([polylinesToAdd isKindOfClass:[NSArray class]]) {
            [_polylinesController addPolylines:polylinesToAdd];
        }
        id polylinesToChange = call.arguments[@"polylinesToChange"];
        if ([polylinesToChange isKindOfClass:[NSArray class]]) {
            [_polylinesController changePolylines:polylinesToChange];
        }
        id polylineIdsToRemove = call.arguments[@"polylineIdsToRemove"];
        if ([polylineIdsToRemove isKindOfClass:[NSArray class]]) {
            [_polylinesController removePolylineIds:polylineIdsToRemove];
        }
        [_mapView centerWithRule:NBCenterMapAroundAll
                         padding:UIEdgeInsetsMake(PADDING, PADDING, PADDING, PADDING)
                        animated:true];
        result(nil);
    } else if ([call.method isEqualToString:@"circles#update"]) {
        id circlesToAdd = call.arguments[@"circlesToAdd"];
        if ([circlesToAdd isKindOfClass:[NSArray class]]) {
            [_circlesController addCircles:circlesToAdd];
        }
        id circlesToChange = call.arguments[@"circlesToChange"];
        if ([circlesToChange isKindOfClass:[NSArray class]]) {
            [_circlesController changeCircles:circlesToChange];
        }
        id circleIdsToRemove = call.arguments[@"circleIdsToRemove"];
        if ([circleIdsToRemove isKindOfClass:[NSArray class]]) {
            [_circlesController removeCircleIds:circleIdsToRemove];
        }
        [_mapView centerWithRule:NBCenterMapAroundAll
                         padding:UIEdgeInsetsMake(PADDING, PADDING, PADDING, PADDING)
                        animated:true];
        result(nil);
    } else if ([call.method isEqualToString:@"map#isCompassEnabled"]) {
        NSNumber* isCompassEnabled = @(_mapView.showsCompass);
        result(isCompassEnabled);
    } else if ([call.method isEqualToString:@"map#isMapToolbarEnabled"]) {
        //TODO: Need confirm, does we need define it.
        NSNumber* isMapToolbarEnabled = [NSNumber numberWithBool:NO];
        result(isMapToolbarEnabled);
    } else if ([call.method isEqualToString:@"map#getMinMaxZoomLevels"]) {
        NSArray* zoomLevels = @[@(_mapView.minZoomLevel), @(_mapView.maxZoomLevel) ];
        result(zoomLevels);
    } else if ([call.method isEqualToString:@"map#isZoomGesturesEnabled"]) {
        NSNumber* isZoomGesturesEnabled = @(_mapView.zoomEnabled);
        result(isZoomGesturesEnabled);
    } else if ([call.method isEqualToString:@"map#isTiltGesturesEnabled"]) {
        NSNumber* isTiltGesturesEnabled = @(_mapView.isRotateCameraEnabled);
        result(isTiltGesturesEnabled);
    } else if ([call.method isEqualToString:@"map#isRotateGesturesEnabled"]) {
        NSNumber* isRotateGesturesEnabled = @(_mapView.isRotateEnabled);
        result(isRotateGesturesEnabled);
    } else if ([call.method isEqualToString:@"map#isScrollGesturesEnabled"]) {
        NSNumber* isScrollGesturesEnabled = @(_mapView.isScrollEnabled);
        result(isScrollGesturesEnabled);
    } else if ([call.method isEqualToString:@"map#isMyLocationButtonEnabled"]) {
        //TODO: AMap does not have the user location button, if use this function, we need manully add to it.
        NSNumber* isMyLocationButtonEnabled = @(false);
        for (UIView* view in _mapView.subviews) {
            if ([view isKindOfClass: [NBUserLocationButton class]]){
                isMyLocationButtonEnabled = @(!view.isHidden);
            }
        }
        result(isMyLocationButtonEnabled);
    } else if ([call.method isEqualToString:@"map#isTrafficEnabled"]) {
        NSNumber* isTrafficEnabled = @(_mapView.isShowTraffic);
        result(isTrafficEnabled);
    } else if ([call.method isEqualToString:@"map#setStyle"]) {
        NSString* mapStyle = [call arguments];
        NSString* error = [self setMapStyle:mapStyle];
        if (error == nil) {
            result(@[ @(YES) ]);
        } else {
            result(@[ @(NO), error ]);
        }
    } else {
        result(FlutterMethodNotImplemented);
    }
}

- (void)showAtX:(CGFloat)x Y:(CGFloat)y {
    _mapView.frame =
    CGRectMake(x, y, CGRectGetWidth(_mapView.frame), CGRectGetHeight(_mapView.frame));
    _mapView.hidden = NO;
}

- (void)hide {
    _mapView.hidden = YES;
}

- (void)updateCamera:(NSArray*)data animate:(BOOL)animate {
    NSString* update = data[0];
    if ([update isEqualToString:@"newCameraPosition"]) {
        [_mapView setMapStatus:ToOptionalMapStatus(data[1]) animated:animate];
        return;
    } else if ([update isEqualToString:@"newLatLng"]) {
        [_mapView setCenterCoordinate:ToLocation(data[1]) animated:animate];
        return;
    } else if ([update isEqualToString:@"newLatLngBounds"]) {
        MACoordinateBounds bounds = ToBounds(data[1]);
        MAMapRect visiableRect = MAMapRectForCoordinateRegion(BoundsToRegion(bounds));
        CGFloat pd = ToFloat(data[2]);
        UIEdgeInsets inset = UIEdgeInsetsMake(pd, pd, pd, pd);
        [_mapView setVisibleMapRect:visiableRect edgePadding:inset animated:animate];
        return;
    } else if ([update isEqualToString:@"newLatLngZoom"]) {
        CGPoint atPivot = [_mapView convertCoordinate:ToLocation(data[1]) toPointToView:_mapView];
        [_mapView setZoomLevel:ToFloat(data[2]) atPivot:atPivot animated:animate];
        return;
    } else if ([update isEqualToString:@"scrollBy"]) {
        CGPoint center = [_mapView convertCoordinate:_mapView.centerCoordinate toPointToView:_mapView];
        CGPoint newCenter = CGPointMake(center.x + ToFloat(data[1]), center.y + ToFloat(data[2]));
        CLLocationCoordinate2D centerCoordinate = [_mapView convertPoint:newCenter toCoordinateFromView:_mapView];
        [_mapView setCenterCoordinate:centerCoordinate animated:false];
        return;
    } else if ([update isEqualToString:@"zoomBy"]) {
        if (data.count == 2) {
            [_mapView setZoomLevel:ToFloat(data[1]) animated:animate];
            return;
        } else {
            CGFloat newZoomLevel = _mapView.zoomLevel * (1 + ToFloat(data[1]));
            if (newZoomLevel <=3.0) { newZoomLevel = 3.0; };
            if (newZoomLevel >=19.0) { newZoomLevel = 19.0; };
            [_mapView setZoomLevel:newZoomLevel atPivot:ToPivot(data[2]) animated:animate];
            return;
        }
    } else if ([update isEqualToString:@"zoomIn"]) {
        if (_mapView.zoomLevel < 3.0) { return; }
        CGFloat newZoomLevel = [_mapView zoomLevel] * (1.0 - ZOOM_RATIO);
        [_mapView setZoomLevel:newZoomLevel animated:animate];
        return;
    } else if ([update isEqualToString:@"zoomOut"]) {
        if (!(_mapView.zoomLevel < 18.9)) { return; }
        CGFloat newZoomLevel = [_mapView zoomLevel] * (1.0 + ZOOM_RATIO);
        [_mapView setZoomLevel:newZoomLevel animated:animate];
        return;
    } else if ([update isEqualToString:@"zoomTo"]) {
        [_mapView setZoomLevel:ToFloat(data[1]) animated:true];
        return;
    } else {
        return ;
    }
}

- (MAMapStatus*)cameraPosition {
    if (_trackCameraPosition) {
        return [_mapView getMapStatus];
    } else {
        return nil;
    }
}

#pragma mark - NBMapOptionsSink methods
- (void)setCameraTargetBounds:(MACoordinateBounds)bounds {
    [_mapView setLimitRegion: BoundsToRegion(bounds)];
}

- (void)setCompassEnabled:(BOOL)enabled {
    _mapView.showsCompass = enabled;
}

- (void)setIndoorEnabled:(BOOL)enabled {
    _mapView.showsIndoorMap = enabled;
    _mapView.showsIndoorMapControl = enabled;
}

- (void)setTrafficEnabled:(BOOL)enabled {
    _mapView.showTraffic = enabled;
}

- (void)setMapType:(MAMapType)mapType {
    _mapView.mapType = mapType;
}

- (void)setMinZoom:(float)minZoom maxZoom:(float)maxZoom {
    [_mapView setMinZoomLevel:minZoom];
    [_mapView setMaxZoomLevel:maxZoom];
}

- (void)setPaddingTop:(float)top left:(float)left bottom:(float)bottom right:(float)right {
    MAMapRect visiableRect = _mapView.visibleMapRect;
    UIEdgeInsets inset = UIEdgeInsetsMake(top, left, bottom, right);
    [_mapView setVisibleMapRect:visiableRect edgePadding:inset animated:false];
}

- (void)setRotateGesturesEnabled:(BOOL)enabled {
    _mapView.rotateEnabled = enabled;
}

- (void)setScrollGesturesEnabled:(BOOL)enabled {
    _mapView.scrollEnabled = enabled;
}

- (void)setTiltGesturesEnabled:(BOOL)enabled {
    _mapView.rotateCameraEnabled = enabled;
}

- (void)setTrackCameraPosition:(BOOL)enabled {
    _trackCameraPosition = enabled;
}

- (void)setZoomGesturesEnabled:(BOOL)enabled {
    _mapView.zoomEnabled = enabled;
}

- (void)setMyLocationEnabled:(BOOL)enabled {
    _mapView.showsUserLocation = enabled;
}

- (void)setMyLocationButtonEnabled:(BOOL)enabled {
    _mapView.showsUserLocation = enabled;
}

- (NSString*)setMapStyle:(NSString*)mapStyle {
    if (mapStyle == (id)[NSNull null] || mapStyle.length == 0) {
        [_mapView setCustomMapStyleEnabled:false];
        return nil;
    }
    NSError* error;
    
    NSString* key = [_registrar lookupKeyForAsset:mapStyle fromPackage:@"nb_map"];
    NSString* path = [[NSBundle mainBundle]pathForResource:key ofType:nil];
    NSData* styleData = [NSData dataWithContentsOfFile:path];
    if (!styleData) {
        return [error localizedDescription];
    } else {
        [_mapView setCustomMapStyleEnabled:true];
        MAMapCustomStyleOptions* options = [[MAMapCustomStyleOptions alloc]init];
        options.styleData = styleData;
        [_mapView setCustomMapStyleOptions:options];
        return nil;
    }
}

#pragma mark - MAMapViewDelegate methods
- (void)mapViewRequireLocationAuth:(CLLocationManager *)locationManager {
     [locationManager requestAlwaysAuthorization];
}

- (void)mapViewDidFinishLoadingMap:(MAMapView *)mapView {
    [_channel invokeMethod:@"camera#onIdle" arguments:@{}];
}
- (void)mapView:(MAMapView *)mapView mapWillMoveByUser:(BOOL)wasUserAction {
    [_channel invokeMethod:@"camera#onMoveStarted" arguments:@{@"isGesture" : @(wasUserAction)}];
}
- (void)mapView:(MAMapView *)mapView mapDidMoveByUser:(BOOL)wasUserAction {
    if (!_cameraDidInitialSetup) {
        _cameraDidInitialSetup = YES;
        [_mapView setMapStatus: [_mapView getMapStatus] animated:true];
    }
    if ((_trackCameraPosition) && [_mapView getMapStatus]) {
        [_channel invokeMethod:@"camera#onMove" arguments:@{@"position" : MapStatusToJson([_mapView getMapStatus])}];
    }
    [_channel invokeMethod:@"camera#onIdle" arguments:@{}];
}

- (void)mapView:(MAMapView *)mapView didTouchPois:(NSArray *)pois {
    //获取地图某个点位的POI的信息
}
- (void)mapView:(MAMapView *)mapView didAnnotationViewTapped:(MAAnnotationView *)view {
    NSString* markerId = ((MAPointAnnotation*)view.annotation).accessibilityLabel;
    [_markersController onMarkerTap:markerId];
}
- (void)mapView:(MAMapView *)mapView annotationView:(MAAnnotationView *)view didChangeDragState:(MAAnnotationViewDragState)newState fromOldState:(MAAnnotationViewDragState)oldState {
    if ((oldState != MAAnnotationViewDragStateEnding) && (newState == MAAnnotationViewDragStateEnding)) {
        NSString* markerId = ((MAPointAnnotation*)view.annotation).accessibilityLabel;
        [_markersController onMarkerTap:markerId];
    }
}

- (void)mapView:(MAMapView *)mapView didAnnotationViewCalloutTapped:(MAAnnotationView *)view {
    NBMarker* marker = (NBMarker*)view.annotation;
    NSString* markerId = marker.accessibilityLabel;
    [_markersController onInfoWindowTap:markerId];
}

- (void)mapView:(MAMapView *)mapView didLongPressedAtCoordinate:(CLLocationCoordinate2D)coordinate {
    [_channel invokeMethod:@"map#onLongPress" arguments:@{@"position" : LocationToJson(coordinate)}];
}

- (void)mapView:(MAMapView *)mapView didSingleTappedAtCoordinate:(CLLocationCoordinate2D)coordinate {
    __block BOOL isHitByOverlay = false;
    [mapView.overlays enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        if ([obj isKindOfClass:[NBMarker class]]){
            //We already have the delegate method handler it,do nothings.
            return;
        }
        if ([obj isKindOfClass:[NBPolygon class]]) {
            NBPolygon* polygon = (NBPolygon*)obj;
            isHitByOverlay = [self isTappedPolygon:polygon atCoordinate:coordinate] && polygon.tappable;
        }
        if ([obj isKindOfClass:[NBPolyline class]]) {
            NBPolyline* polyline = (NBPolyline*)obj;
            isHitByOverlay = [self isTappedPolyline:polyline atCoordinate:coordinate] && polyline.tappable;
        }
        if ([obj isKindOfClass:[NBCircle class]]) {
            NBCircle* circle = (NBCircle*)obj;
            isHitByOverlay = [self isTappedCircle:circle atCoordinate:coordinate] && circle.tappable;
        }
        *stop = isHitByOverlay;
    }];
    if (isHitByOverlay) { return; }
    [_channel invokeMethod:@"map#onTap" arguments:@{@"position" : LocationToJson(coordinate)}];
}

#pragma mark - annotationViews
- (MAAnnotationView *)mapView:(MAMapView *)mapView viewForAnnotation:(id<MAAnnotation>)annotation {
    if ([annotation isKindOfClass:[MAUserLocation class]]) {
        MAUserLocation* userLocation = (MAUserLocation*)annotation;
        return [NBMarkerViewsFactory createUserAnnotationViewWith:userLocation mapView:(MAMapView *)_mapView];
    }
    if ([annotation isKindOfClass:[NBMarker class]]) {
        NBMarker* marker = (NBMarker*)annotation;
        return [NBMarkerViewsFactory createAnnotationViewWith:marker mapView:_mapView];
    }
    return nil;
}

- (MAOverlayRenderer *)mapView:(MAMapView *)mapView rendererForOverlay:(id<MAOverlay>)overlay {
    if ([overlay isKindOfClass:[NBPolyline class]]){
        NBPolyline* polyline = (NBPolyline*)overlay;
        return [NBRendersFactory createWithPolyline:polyline];
    }
    if ([overlay isKindOfClass:[NBPolygon class]]){
        NBPolygon* polygon = (NBPolygon*)overlay;
        return [NBRendersFactory createWithPolygon:polygon];
    }
    if ([overlay isKindOfClass:[NBCircle class]]){
        NBCircle* circle = (NBCircle*)overlay;
        return [NBRendersFactory createWithCircle:circle];
    }
    return nil;
}

#pragma mark - Help Methods
- (BOOL)isTappedCircle:(NBCircle*)circle atCoordinate:(CLLocationCoordinate2D)coordinate{
    BOOL isHit = false;
    MAMapPoint tappedCenter = MAMapPointForCoordinate(coordinate);
    MAMapPoint circleCenter = MAMapPointForCoordinate(circle.coordinate);
    isHit = MACircleContainsPoint(tappedCenter, circleCenter, circle.radius);
    return isHit;
}
- (BOOL)isTappedPolygon:(NBPolygon*)polygon atCoordinate:(CLLocationCoordinate2D)coordinate {
    BOOL isHit = false;
    MAMapPoint tappedCenter = MAMapPointForCoordinate(coordinate);
    isHit = MAPolygonContainsPoint(tappedCenter, polygon.points, polygon.pointCount);
    return isHit;
}

- (BOOL)isTappedPolyline:(NBPolyline*)polyline atCoordinate:(CLLocationCoordinate2D)coordinate {
    BOOL isHit = false;
    isHit = [NBCommonUtility polylineHitTestWithCoordinate:coordinate
                                                   mapView:_mapView
                                            polylinePoints:polyline.points
                                                pointCount:polyline.pointCount
                                                 lineWidth:polyline.lineWidth ];
    return isHit;
}

@end

@implementation NBUserLocationButton

- (instancetype)init {
    self = [super init];
    if (self){
        self.frame = CGRectMake(0.0, 0.0, 40.0, 40.0);
        self.backgroundColor = UIColor.grayColor;
        self.layer.cornerRadius = 5.0;
        self.layer.borderWidth = 1.0;
        self.layer.shadowOffset = CGSizeMake(2.0, 2.0);
        self.layer.shadowColor = UIColor.blackColor.CGColor;
        self.layer.masksToBounds = true;
    }
    return self;
}

@end

#pragma mark - Implementations of JSON conversion functions.
static NSDictionary* PointToJson(CGPoint point) {
    return @{
             @"x" : @((int)point.x),
             @"y" : @((int)point.y),
             };
}
static NSArray* LocationToJson(CLLocationCoordinate2D position){
    return @[ @(position.latitude), @(position.longitude) ];
}
static CLLocationCoordinate2D ToLocation(NSArray* data) {
    return [NBJsonConversions toLocation:data];
}
static NSDictionary* MapStatusToJson(MAMapStatus* mapStatus) {
    if (!mapStatus) {
        return nil;
    }
    return @{
             @"target" : LocationToJson([mapStatus centerCoordinate]),
             @"zoom" : @([mapStatus zoomLevel]),
             @"bearing" : @([mapStatus rotationDegree]),
             @"tilt" : @([mapStatus cameraDegree]),
             };
}
static CGPoint ToCGPoint(NSDictionary* data) {
    double x = ToDouble(data[@"x"]);
    double y = ToDouble(data[@"y"]);
    return CGPointMake(x, y);
}
static CGPoint ToPivot(NSArray* data) {
    double x = ToDouble(data[0]);
    double y = ToDouble(data[0]);
    return CGPointMake(x, y);
}
static MAMapStatus* ToMapStatus(NSDictionary* data) {
    return [[MAMapStatus alloc]initWithCenterCoordinate:ToLocation(data[@"target"])
                                              zoomLevel:ToFloat(data[@"zoom"])
                                         rotationDegree:ToDouble(data[@"bearing"])
                                           cameraDegree:ToDouble(data[@"tilt"])
                                           screenAnchor: CGPointMake(0.5, 0.5)];
}
static MAMapStatus* ToOptionalMapStatus(NSDictionary* data) {
    return data ? ToMapStatus(data) : nil;
}
static MACoordinateBounds ToBounds(NSArray* data) {
    return MACoordinateBoundsMake(ToLocation(data[1]), ToLocation(data[0]));
}
static MACoordinateBounds ToOptionalBounds(NSArray* data) {
    CLLocationCoordinate2D northEast = CLLocationCoordinate2DMake(90, 360.0);
    CLLocationCoordinate2D southWest = CLLocationCoordinate2DMake(-90, 360.0);
    MACoordinateBounds defaultBounds = MACoordinateBoundsMake(northEast, southWest);
    return ([data[0] isKindOfClass:[NSNull class]]) ? defaultBounds : ToBounds(data[0]);
}

static MACoordinateBounds RegionToBounds(MACoordinateRegion region) {
    MACoordinateSpan span = region.span;
    double lat = region.center.latitude;
    double lon = region.center.longitude;
    //注意: 中国🇨🇳区域属于南半球,纬度计算与精度计算相反
    CLLocationCoordinate2D northEast = CLLocationCoordinate2DMake(lat + span.latitudeDelta, lon + span.longitudeDelta);
    CLLocationCoordinate2D southWest = CLLocationCoordinate2DMake(lat - span.latitudeDelta, lon - span.longitudeDelta);
    return MACoordinateBoundsMake(northEast, southWest);
}

static MACoordinateRegion BoundsToRegion(MACoordinateBounds bounds) {
    CLLocationCoordinate2D ne = bounds.northEast;
    CLLocationCoordinate2D sw = bounds.southWest;
    CLLocationDegrees latitude = (ne.latitude + sw.latitude) * 0.5;
    CLLocationDegrees longitude = (ne.longitude + sw.longitude) * 0.5;
    CLLocationCoordinate2D center = CLLocationCoordinate2DMake(latitude, longitude);
    CLLocationDegrees spanY = ne.latitude - center.latitude;
    CLLocationDegrees spanX = ne.longitude - center.longitude;
    MACoordinateSpan span = MACoordinateSpanMake(spanY, spanX);
    return MACoordinateRegionMake(center, span);
}
static NSDictionary* MACoordinateBoundsToJson(MACoordinateBounds bounds) {
    return @{
             @"southwest" : LocationToJson(bounds.southWest),
             @"northeast" : LocationToJson(bounds.northEast),
             };
}
static double ToDouble(NSNumber* data) { return [NBJsonConversions toDouble:data];}

static float ToFloat(NSNumber* data) { return [NBJsonConversions toFloat:data]; }


static int ToInt(NSNumber* data) { return [NBJsonConversions toInt:data]; }

static BOOL ToBool(NSNumber* data) { return [NBJsonConversions toBool:data]; }

static MAMapType ToMapViewType(NSNumber* json) {
    int mapTypeValue = ToInt(json);
    switch (mapTypeValue) {
        case 0: return MAMapTypeStandard;
        case 1: return MAMapTypeSatellite;
        case 2: return MAMapTypeStandardNight;
        case 3: return MAMapTypeNavi;
        case 4: return MAMapTypeBus;
        default:  return MAMapTypeStandard;
    }
}

static void InterpretMapOptions(NSDictionary* data, id<NBMapOptionsSink> sink) {
    NSArray* cameraTargetBounds = data[@"cameraTargetBounds"];
    if (cameraTargetBounds && ![cameraTargetBounds[0] isKindOfClass:[NSNull class]] ) {
        [sink setCameraTargetBounds:ToOptionalBounds(cameraTargetBounds)];
    }
    NSNumber* compassEnabled = data[@"compassEnabled"];
    if (compassEnabled) {
        [sink setCompassEnabled:ToBool(compassEnabled)];
    }
    id indoorEnabled = data[@"indoorEnabled"];
    if (indoorEnabled) {
        [sink setIndoorEnabled:ToBool(indoorEnabled)];
    }
    id trafficEnabled = data[@"trafficEnabled"];
    if (trafficEnabled) {
        [sink setTrafficEnabled:ToBool(trafficEnabled)];
    }
    id mapType = data[@"mapType"];
    if (mapType) {
        [sink setMapType:ToMapViewType(mapType)];
    }
    NSArray* zoomData = data[@"minMaxZoomPreference"];
    if (zoomData && zoomData[0] != [NSNull null] && zoomData[1] != [NSNull null] ) {
        float minZoom = ToFloat(zoomData[0]);
        float maxZoom = ToFloat(zoomData[1]);
        [sink setMinZoom:minZoom maxZoom:maxZoom];
    }
    NSArray* paddingData = data[@"padding"];
    if (paddingData) {
        float top = (paddingData[0] == [NSNull null]) ? 0 : ToFloat(paddingData[0]);
        float left = (paddingData[1] == [NSNull null]) ? 0 : ToFloat(paddingData[1]);
        float bottom = (paddingData[2] == [NSNull null]) ? 0 : ToFloat(paddingData[2]);
        float right = (paddingData[3] == [NSNull null]) ? 0 : ToFloat(paddingData[3]);
        if (!(top == 0 && left == 0 && bottom == 0 && right == 0)) {
            [sink setPaddingTop:top left:left bottom:bottom right:right];
        }
    }
    
    NSNumber* rotateGesturesEnabled = data[@"rotateGesturesEnabled"];
    if (rotateGesturesEnabled) {
        [sink setRotateGesturesEnabled:ToBool(rotateGesturesEnabled)];
    }
    NSNumber* scrollGesturesEnabled = data[@"scrollGesturesEnabled"];
    if (scrollGesturesEnabled) {
        [sink setScrollGesturesEnabled:ToBool(scrollGesturesEnabled)];
    }
    NSNumber* tiltGesturesEnabled = data[@"tiltGesturesEnabled"];
    if (tiltGesturesEnabled) {
        [sink setTiltGesturesEnabled:ToBool(tiltGesturesEnabled)];
    }
    NSNumber* trackCameraPosition = data[@"trackCameraPosition"];
    if (trackCameraPosition) {
        [sink setTrackCameraPosition:ToBool(trackCameraPosition)];
    }
    NSNumber* zoomGesturesEnabled = data[@"zoomGesturesEnabled"];
    if (zoomGesturesEnabled) {
        [sink setZoomGesturesEnabled:ToBool(zoomGesturesEnabled)];
    }
    NSNumber* myLocationEnabled = data[@"myLocationEnabled"];
    if (myLocationEnabled) {
        [sink setMyLocationEnabled:ToBool(myLocationEnabled)];
    }
    NSNumber* myLocationButtonEnabled = data[@"myLocationButtonEnabled"];
    if (myLocationButtonEnabled) {
        [sink setMyLocationButtonEnabled:ToBool(myLocationButtonEnabled)];
    }
}

