//
//  NBMapMarkerController.m
//  nb_map
//
//  Created by Guo Longxiang, (Longxiang.Guo@partner.bmw.com) on 2019/10/29.
//

#import "NBMapMarkerController.h"
#import "NBJsonConversions.h"
#import "MapPlugin.h"
#import "MAMapView+CenterAroundVisiableOverlays.h"
#import "NBMarkerViewsFactory.h"

static UIImage* ExtractIcon(NSObject<FlutterPluginRegistrar>* registrar, NSArray* icon);
static UIImage* defaultMarkIcon(void);
static void InterpretInfoWindow(id<NBMapMarkerOptionsSink> sink, NSDictionary* data);

@implementation NBMapMarkerController {
    NBMarker* _marker;
    MAAnnotationView* _markerView;
    MAMapView* _mapView;
}

- (instancetype)initMarkerWithPosition:(CLLocationCoordinate2D)position
                              markerId:(NSString*)markerId
                               mapView:(MAMapView*)mapView {
    self = [super init];
    if (self) {
        _marker = [[NBMarker alloc]init];
        _mapView = mapView;
        _markerId = markerId;
        _marker.coordinate = position;
        _marker.accessibilityLabel = markerId;
        _marker.visiable = true;
        _marker.tappable = true;
        _marker.draggable = false;
        _marker.anchor = CGPointMake(0.5, 0.5);
        _marker.windowInfoAnchor = CGPointMake(0.5, 0.5);
        _marker.rotation = 0.0;
        _marker.icon = defaultMarkIcon();
        _marker.canShowCallOut = false;
        _marker.canShowAccessoryView = false;
        _shouldConsumeTapEvents = true;
    }
    return self;
}

- (MAAnnotationView*)annotationView {
    if (_markerView == nil){
        _markerView = [_mapView viewForAnnotation:_marker];
    }
    return _markerView;
}

- (void)addMarker {
    [_mapView addAnnotation:_marker];
    if (_marker.isVisiable){
        [_mapView updateCachedMarker:_marker event:NBVisiableOverlayAdd];
    }
}

- (void)removeMarker {
    [_mapView removeAnnotation:_marker];
    _markerView = nil;
    [_mapView updateCachedMarker:_marker event:NBVisiableOverlayDelete];
}

#pragma mark - FLTGoogleMapMarkerOptionsSink methods

- (void)setAlpha:(float)alpha {
    _marker.alpha = alpha;
    if (![self annotationView]){
        return;
    }
    [self annotationView].alpha = alpha;
}
- (void)setAnchor:(CGPoint)anchor {
    _marker.anchor = anchor;
    if (![self annotationView]) {
        return;
    }
    CGSize size = [self annotationView].bounds.size;
    [self annotationView].centerOffset = CGPointMake(size.width*anchor.x, size.height*anchor.y);
    
}

- (void)setInfoWindowAnchor:(CGPoint)anchor {
    _marker.anchor = anchor;
    if (![self annotationView]) {
        return;
    }
    CGSize size = [self annotationView].bounds.size;
    [self annotationView].calloutOffset = CGPointMake(size.width*anchor.x, size.height*anchor.y);
}
- (void)setInfoWindowTitle:(NSString*)title snippet:(NSString*)snippet {
    _marker.title = title;
    _marker.subtitle = snippet;
    if (![self annotationView]) {
        return;
    }
    //TODO: updateCallView;
    [self annotationView].customCalloutView = [[MACustomCalloutView alloc]init];
}

- (void)setConsumeTapEvents:(BOOL)consumes {
    _marker.tappable = consumes;
    _shouldConsumeTapEvents = consumes;
    if (![self annotationView]) {
        return;
    }
    [self annotationView].userInteractionEnabled = consumes;
}

- (void)setDraggable:(BOOL)draggable {
    _marker.draggable = draggable;
    if (![self annotationView]){
        return;
    }
    [self annotationView].draggable = draggable;
}
- (void)setIcon:(UIImage*)icon {
    _marker.icon = icon;
    if (![self annotationView]){
        return;
    }
    [self annotationView].image = [NBMarkerViewsFactory createMarkerViewWith:_marker];
}

- (void)setPosition:(CLLocationCoordinate2D)position {
    [_marker setCoordinate:position];
}
- (void)setRotation:(CLLocationDegrees)rotation {
    _marker.rotation = rotation;
    if (![self annotationView]){
        return;
    }
    [self annotationView].image = [NBMarkerViewsFactory createMarkerViewWith:_marker];
}
- (void)setVisible:(BOOL)visible {
    _marker.visiable = visible;
    if (_marker.isVisiable){
        [_mapView updateCachedMarker:_marker event:NBVisiableOverlayAdd];
    } else {
        [_mapView updateCachedMarker:_marker event:NBVisiableOverlayDelete];
    }
    if (![self annotationView]){
        return;
    }
    [self annotationView].hidden = !visible;
}
- (void)setShowCallOut:(BOOL)showCallOut {
    _marker.canShowCallOut = showCallOut;
    if (![self annotationView]){
        return;
    }
    [self annotationView].canShowCallout = showCallOut;
}
- (void)setZIndex:(int)zIndex {
    if (zIndex >= 1) {
        _marker.level = MAOverlayLevelAboveLabels;
    } else {
        _marker.level = MAOverlayLevelAboveRoads;
    }
}

- (void)setNeedsUpdate {
    [[self annotationView] setNeedsDisplay];
    [[self annotationView] setNeedsLayout];
}

@end

static double ToDouble(NSNumber* data) { return [NBJsonConversions toDouble:data]; }

static float ToFloat(NSNumber* data) { return [NBJsonConversions toFloat:data]; }

static CLLocationCoordinate2D ToLocation(NSArray* data) {
    return [NBJsonConversions toLocation:data];
}

static int ToInt(NSNumber* data) { return [NBJsonConversions toInt:data]; }

static BOOL ToBool(NSNumber* data) { return [NBJsonConversions toBool:data]; }

static CGPoint ToPoint(NSArray* data) { return [NBJsonConversions toPoint:data]; }

static NSArray* PositionToJson(CLLocationCoordinate2D data) {
    return [NBJsonConversions positionToJson:data];
}

static void InterpretMarkerOptions(NSDictionary* data, id<NBMapMarkerOptionsSink> sink,
                                   NSObject<FlutterPluginRegistrar>* registrar) {
    NSNumber* alpha = data[@"alpha"];
    if (alpha) {
        [sink setAlpha:ToFloat(alpha)];
    }
    NSArray* anchor = data[@"anchor"];
    if (anchor) {
        [sink setAnchor:ToPoint(anchor)];
    }
    NSNumber* draggable = data[@"draggable"];
    if (draggable) {
        [sink setDraggable:ToBool(draggable)];
    }
    NSArray* icon = data[@"icon"];
    if (icon) {
        UIImage* image = ExtractIcon(registrar, icon);
        [sink setIcon:image];
    }
    NSNumber* consumeTapEvents = data[@"consumeTapEvents"];
    if (consumeTapEvents) {
        [sink setConsumeTapEvents:ToBool(consumeTapEvents)];
    }
    InterpretInfoWindow(sink, data);
    NSArray* position = data[@"position"];
    if (position) {
        [sink setPosition:ToLocation(position)];
    }
    NSNumber* rotation = data[@"rotation"];
    if (rotation) {
        [sink setRotation:ToDouble(rotation)];
    }
    NSNumber* visible = data[@"visible"];
    if (visible) {
        [sink setVisible:ToBool(visible)];
    }
    NSNumber* zIndex = data[@"zIndex"];
    if (zIndex) {
        [sink setZIndex:ToInt(zIndex)];
    }
}

static void InterpretInfoWindow(id<NBMapMarkerOptionsSink> sink, NSDictionary* data) {
    NSDictionary* infoWindow = data[@"infoWindow"];
    if (infoWindow) {
        NSString* title = infoWindow[@"title"];
        NSString* snippet = infoWindow[@"snippet"];
        if (title) {
            [sink setInfoWindowTitle:title snippet:snippet];
        }
        NSArray* infoWindowAnchor = infoWindow[@"infoWindowAnchor"];
        if (infoWindowAnchor) {
            [sink setInfoWindowAnchor:ToPoint(infoWindowAnchor)];
        }
    }
}

static UIImage* defaultMarkIcon(void) {
    NSString* resourcePath = [[NBMapPlugin registrar] lookupKeyForAsset:@"images/default_marker.png" fromPackage:@"nb_map"];
    NSString* imagePath = [NSString stringWithFormat:@"%@/%@", [NSBundle mainBundle].bundlePath, resourcePath];
    UIImage* image = [UIImage imageWithContentsOfFile:imagePath];
    return image;
}

static UIImage* scaleImage(UIImage* image, NSNumber* scaleParam) {
    double scale = 1.0;
    if ([scaleParam isKindOfClass:[NSNumber class]]) {
        scale = scaleParam.doubleValue;
    }
    if (fabs(scale - 1) > 1e-3) {
        return [UIImage imageWithCGImage:[image CGImage]
                                   scale:(image.scale * scale)
                             orientation:(image.imageOrientation)];
    }
    return image;
}

static UIImage* ExtractIcon(NSObject<FlutterPluginRegistrar>* registrar, NSArray* iconData) {
    UIImage* image;
    if ([iconData.firstObject isEqualToString:@"defaultMarker"]) {
        image = defaultMarkIcon();

    } else if ([iconData.firstObject isEqualToString:@"fromAsset"]) {
        if (iconData.count == 2) {
            image = [UIImage imageNamed:[registrar lookupKeyForAsset:iconData[1]]];
        } else {
            image = [UIImage imageNamed:[registrar lookupKeyForAsset:iconData[1]
                                                         fromPackage:iconData[2]]];
        }
    } else if ([iconData.firstObject isEqualToString:@"fromAssetImage"]) {
        if (iconData.count == 3) {
            image = [UIImage imageNamed:[registrar lookupKeyForAsset:iconData[1]]];
            NSNumber* scaleParam = iconData[2];
            image = scaleImage(image, scaleParam);
        } else {
            NSString* error =
            [NSString stringWithFormat:@"'fromAssetImage' should have exactly 3 arguments. Got: %lu",
             (unsigned long)iconData.count];
            NSException* exception = [NSException exceptionWithName:@"InvalidBitmapDescriptor"
                                                             reason:error
                                                           userInfo:nil];
            @throw exception;
        }
    } else if ([iconData[0] isEqualToString:@"fromBytes"]) {
        if (iconData.count == 2) {
            @try {
                FlutterStandardTypedData* byteData = iconData[1];
                CGFloat screenScale = [[UIScreen mainScreen] scale];
                image = [UIImage imageWithData:[byteData data] scale:screenScale];
            } @catch (NSException* exception) {
                @throw [NSException exceptionWithName:@"InvalidByteDescriptor"
                                               reason:@"Unable to interpret bytes as a valid image."
                                             userInfo:nil];
            }
        } else {
            NSString* error = [NSString
                               stringWithFormat:@"fromBytes should have exactly one argument, t%lubytes.",
                               (unsigned long)iconData.count];
            NSException* exception = [NSException exceptionWithName:@"InvalidByteDescriptor"
                                                             reason:error
                                                           userInfo:nil];
            @throw exception;
        }
    }
    
    return image;
}

@implementation NBMapMarkersController {
    NSMutableDictionary* _markerIdToController;
    FlutterMethodChannel* _methodChannel;
    NSObject<FlutterPluginRegistrar>* _registrar;
    MAMapView* _mapView;
}
- (instancetype)init:(FlutterMethodChannel*)methodChannel
             mapView:(MAMapView*)mapView
           registrar:(NSObject<FlutterPluginRegistrar>*)registrar {
    self = [super init];
    if (self) {
        _methodChannel = methodChannel;
        _mapView = mapView;
        _markerIdToController = [NSMutableDictionary dictionaryWithCapacity:1];
        _registrar = registrar;
    }
    return self;
}
- (void)addMarkers:(NSArray*)markersToAdd {
    for (NSDictionary* marker in markersToAdd) {
        CLLocationCoordinate2D position = [NBMapMarkersController getPosition:marker];
        NSString* markerId = [NBMapMarkersController getMarkerId:marker];
        NBMapMarkerController* controller =
        [[NBMapMarkerController alloc] initMarkerWithPosition:position
                                                            markerId:markerId
                                                             mapView:_mapView];
        InterpretMarkerOptions(marker, controller, _registrar);
        _markerIdToController[markerId] = controller;
        [controller addMarker];
    }
}
- (void)changeMarkers:(NSArray*)markersToChange {
    for (NSDictionary* marker in markersToChange) {
        NSString* markerId = [NBMapMarkersController getMarkerId:marker];
        NBMapMarkerController* controller = _markerIdToController[markerId];
        if (!controller) {
            continue;
        }
        InterpretMarkerOptions(marker, controller, _registrar);
    }
}
- (void)removeMarkerIds:(NSArray*)markerIdsToRemove {
    for (NSString* markerId in markerIdsToRemove) {
        if (!markerId) {
            continue;
        }
        NBMapMarkerController* controller = _markerIdToController[markerId];
        if (!controller) {
            continue;
        }
        [controller removeMarker];
        [_markerIdToController removeObjectForKey:markerId];
    }
}
- (BOOL)onMarkerTap:(NSString*)markerId {
    if (!markerId) {
        return NO;
    }
    NBMapMarkerController* controller = _markerIdToController[markerId];
    if (!controller) {
        return NO;
    }
    [_methodChannel invokeMethod:@"marker#onTap" arguments:@{@"markerId" : markerId}];
    return controller.shouldConsumeTapEvents;
}
- (void)onMarkerDragEnd:(NSString*)markerId coordinate:(CLLocationCoordinate2D)coordinate {
    if (!markerId) {
        return;
    }
    NBMapMarkerController* controller = _markerIdToController[markerId];
    if (!controller) {
        return;
    }
    [_methodChannel invokeMethod:@"marker#onDragEnd"
                       arguments:@{@"markerId" : markerId, @"position" : PositionToJson(coordinate)}];
}

- (void)onInfoWindowTap:(NSString*)markerId {
    if (markerId && _markerIdToController[markerId]) {
        [_methodChannel invokeMethod:@"infoWindow#onTap" arguments:@{@"markerId" : markerId}];
    }
}

+ (CLLocationCoordinate2D)getPosition:(NSDictionary*)marker {
    NSArray* position = marker[@"position"];
    return ToLocation(position);
}

+ (NSString*)getMarkerId:(NSDictionary*)marker {
    return marker[@"markerId"];
}

@end

@implementation NBMarker

- (BOOL)isEqual:(id)object {
    if (![object isKindOfClass:[NBMarker class]]){
        return false;
    }
    NBMarker* lhs = (NBMarker*)object;
    BOOL superJudgement = [super isEqual:object];
    BOOL childJudgement = self.accessibilityLabel == lhs.accessibilityLabel;
    return superJudgement && childJudgement;
    
}

@end
