//
//  NBMapPolygonController.m
//  nb_map
//
//  Created by Guo Longxiang, (Longxiang.Guo@partner.bmw.com) on 2019/10/29.
//

#import "NBMapPolygonController.h"
#import "NBJsonConversions.h"
#import "MAMapView+CenterAroundVisiableOverlays.h"

@implementation NBMapPloygonController{
    MAMapView* _mapView;
    NBPolygon* _polygon;
    MAPolygonRenderer* _polygonRenderer;
}

- (instancetype)initPolygonWithCoord:(CLLocationCoordinate2D*)coords
                           polygonId:(NSString*)polygonId
                             mapView:(MAMapView*)mapView {
    self = [super init];
    if (self) {
        NSInteger count = sizeof(coords)/sizeof(coords[0]);
        _polygon = [NBPolygon polygonWithCoordinates:coords count:count];
      
        
        _polygon.accessibilityLabel = polygonId;
        _polygon.tappable = false;
        _polygon.visible = true;
        _polygon.fillColor = kMAOverlayRendererDefaultFillColor;
        _polygon.strokeColor = kMAOverlayRendererDefaultStrokeColor;
        _polygon.level = MAOverlayLevelAboveLabels;
        _polygon.lineWidth = 0;
        _polygonId = polygonId;
        _mapView = mapView;
    }
    return self;
}

#pragma mark - Help Methods
- (MAPolygonRenderer*) polygonRender {
    if (_polygonRenderer == nil) {
        _polygonRenderer = (MAPolygonRenderer*)[_mapView rendererForOverlay:_polygon];
    }
    return _polygonRenderer;
}

- (void)addPolygon {
    [_mapView addOverlay:_polygon level:_polygon.level];
    if(_polygon.isVisiable){
        [_mapView updateCachedOverlay:_polygon event:NBVisiableOverlayAdd];
    }
}

- (void)removePolygon {
    [_mapView removeOverlay:_polygon];
    _polygonRenderer = nil;
    [_mapView updateCachedOverlay:_polygon event:NBVisiableOverlayDelete];
}

#pragma mark - <NBMapPolygonOptionsSink>
- (void)setConsumeTapEvents:(BOOL)consume {
    _polygon.tappable = consume;
}

- (void)setVisible:(BOOL)visible {
    if (_polygon.isVisiable){
        [_mapView updateCachedOverlay:_polygon event:NBVisiableOverlayAdd];
    } else{
        [_mapView updateCachedOverlay:_polygon event:NBVisiableOverlayDelete];
    }
    if (_polygon.visible == visible) {
        return;
    }
    if (![self polygonRender]) {
        return;
    }
    CGFloat alpha = visible ? 1.0 : 0.0;
    UIColor* newFillColor = [[_polygonRenderer.fillColor copy]colorWithAlphaComponent:alpha];
    UIColor* newStrokeColor = [[_polygonRenderer.strokeColor copy]colorWithAlphaComponent:alpha];
    _polygonRenderer.fillColor = newFillColor;
    _polygonRenderer.strokeColor = newStrokeColor;
    _polygon.fillColor = newFillColor;
    _polygon.strokeColor = newStrokeColor;
    _polygon.visible = visible;
}

- (void)setFillColor:(UIColor*)color {
    _polygon.fillColor = color;
    if(![self polygonRender]){
        return;
    }
    [self polygonRender].fillColor = color;
}

- (void)setStrokeColor:(UIColor*)color {
    _polygon.strokeColor = color;
    if(![self polygonRender]){
        return;
    }
    [self polygonRender].strokeColor = color;
    
}

- (void)setStrokeWidth:(CGFloat)width {
    _polygon.lineWidth = width;
    if(![self polygonRender]){
        return;
    }
    [self polygonRender].lineWidth = width;
}

- (void)setPoints:(NSArray<CLLocation*>*)points {
    CLLocationCoordinate2D coords[points.count];
    for (int index = 0; index < points.count; index++) {
        CLLocation* location = points[index];
        coords[index] = location.coordinate;
    }
    [_polygon setPolygonWithCoordinates:coords count:points.count];
}

- (void)setZIndex:(int)zIndex {
    if (zIndex >= 1) {
        _polygon.level = MAOverlayLevelAboveLabels;
    } else {
        _polygon.level = MAOverlayLevelAboveRoads;
    }
}

- (void)setNeedsUpdate {
    if (![self polygonRender]) {
        return;
    }
    [[self polygonRender] setNeedsUpdate];
}
@end


static int ToInt(NSNumber* data) { return [NBJsonConversions toInt:data]; }

static BOOL ToBool(NSNumber* data) { return [NBJsonConversions toBool:data]; }

static NSArray<CLLocation*>* ToPoints(NSArray* data) {
    return [NBJsonConversions toPoints:data];
}

static UIColor* ToColor(NSNumber* data) { return [NBJsonConversions toColor:data]; }

static void InterpretPolygonOptions(NSDictionary* data, id<NBMapPolygonOptionsSink> sink,
                                    NSObject<FlutterPluginRegistrar>* registrar) {
    NSNumber* consumeTapEvents = data[@"consumeTapEvents"];
    if (consumeTapEvents) {
        [sink setConsumeTapEvents:ToBool(consumeTapEvents)];
    }
    
    NSNumber* visible = data[@"visible"];
    if (visible) {
        [sink setVisible:ToBool(visible)];
    }
    
    NSNumber* zIndex = data[@"zIndex"];
    if (zIndex) {
        [sink setZIndex:ToInt(zIndex)];
    }
    
    NSArray* points = data[@"points"];
    if (points) {
        [sink setPoints:ToPoints(points)];
    }
    
    NSNumber* fillColor = data[@"fillColor"];
    if (fillColor) {
        [sink setFillColor:ToColor(fillColor)];
    }
    
    NSNumber* strokeColor = data[@"strokeColor"];
    if (strokeColor) {
        [sink setStrokeColor:ToColor(strokeColor)];
    }
    
    NSNumber* strokeWidth = data[@"strokeWidth"];
    if (strokeWidth) {
        [sink setStrokeWidth:ToInt(strokeWidth)];
    }
    [sink setNeedsUpdate];
}

@implementation NBMapPolygonsController {
    NSMutableDictionary* _polygonIdToController;
    FlutterMethodChannel* _methodChannel;
    NSObject<FlutterPluginRegistrar>* _registrar;
    MAMapView* _mapView;
}

- (instancetype)init:(FlutterMethodChannel*)methodChannel
             mapView:(MAMapView*)mapView
           registrar:(NSObject<FlutterPluginRegistrar>*)registrar {
    self = [super init];
    if (self) {
        _methodChannel = methodChannel;
        _mapView = mapView;
        _polygonIdToController = [NSMutableDictionary dictionaryWithCapacity:1];
        _registrar = registrar;
    }
    return self;
}

- (void)addPolygons:(NSArray*)polygonsToAdd {
    for (NSDictionary* polygon in polygonsToAdd) {
        NSArray<CLLocation*>* locations = ToPoints(polygon[@"points"]);
        CLLocationCoordinate2D coords[locations.count];
        [NBMapPolygonsController getCoords:locations coords:coords];
        NSString* polygonId = [NBMapPolygonsController getPolygonId: polygon];
        NBMapPloygonController* controller = [[NBMapPloygonController alloc]initPolygonWithCoord:coords polygonId:polygonId mapView:_mapView];
        InterpretPolygonOptions(polygon, controller, _registrar);
        _polygonIdToController[polygonId] = controller;
        [controller addPolygon];
    }
}

- (void)changePolygons:(NSArray*)polygonsToChange {
    for (NSDictionary* polygon in polygonsToChange) {
        NSString* polygonId = [NBMapPolygonsController getPolygonId:polygon];
        NBMapPloygonController* controller = _polygonIdToController[polygonId];
        if (!controller) {
            continue;
        }
        InterpretPolygonOptions(polygon, controller, _registrar);
    }
}

- (void)removePolygonIds:(NSArray*)polygonIdsToRemove {
    for (NSString* polygonId in polygonIdsToRemove) {
        if (!polygonId) {
            continue;
        }
        NBMapPloygonController* controller = _polygonIdToController[polygonId];
        if (!controller) {
            continue;
        }
        [controller removePolygon];
        [_polygonIdToController removeObjectForKey:polygonId];
    }
}

- (void)onPolygonTap:(NSString*)polygonId {
    if (!polygonId) {
        return;
    }
    NBMapPloygonController* controller = _polygonIdToController[polygonId];
    if (!controller) {
        return;
    }
    [_methodChannel invokeMethod:@"polygon#onTap" arguments:@{@"polygonId" : polygonId}];
}

- (bool)hasPolygonWithId:(NSString*)polygonId {
    if (!polygonId) {
        return false;
    }
    return _polygonIdToController[polygonId] != nil;
}
+ (void)getCoords:(NSArray<CLLocation*>*) points coords:(CLLocationCoordinate2D*)coords {
    for (NSUInteger index = 0; index < points.count; index++) {
        coords[index] = points[index].coordinate;
    }
}
+ (NSString*)getPolygonId:(NSDictionary*)polygon {
    return polygon[@"polygonId"];
}
@end

@implementation NBPolygon

@end
