//
//  NBMapPloylineController.m
//  nb_map
//
//  Created by Guo Longxiang, (Longxiang.Guo@partner.bmw.com) on 2019/10/29.
//

#import "NBMapPolylineController.h"
#import "NBJsonConversions.h"
#import "MAMapView+CenterAroundVisiableOverlays.h"

@implementation NBMapPolylineController{
    NBPolyline* _polyline;
    MAMapView* _mapView;
    MAPolylineRenderer* _polylineRenderer;
}

- (instancetype)initPolylineWithCoords:(CLLocationCoordinate2D*)coords
                                 count:(NSUInteger)count
                            polylineId:(NSString*)polylineId
                               mapView:(MAMapView*)mapView {
    self = [super init];
    if (self) {
        _polyline = [NBPolyline polylineWithCoordinates:coords count: count];
        _mapView = mapView;
        _polylineId = polylineId;
        _polyline.accessibilityLabel = polylineId;
        _polyline.tappable = false;
        _polyline.visible = true;
        _polyline.level = MAOverlayLevelAboveRoads;
        _polyline.fillColor = kMAOverlayRendererDefaultFillColor;
        _polyline.strokeColor = kMAOverlayRendererDefaultStrokeColor;
        _polyline.lineWidth = 0;
    }
    return self;
}

#pragma mark - Help Methods
- (MAPolylineRenderer*)polylineRenderer {
    if (_polylineRenderer == nil){
        _polylineRenderer = (MAPolylineRenderer*)[_mapView rendererForOverlay:_polyline];
    }
    return _polylineRenderer;
}

- (void)addPolyine {
    [_mapView addOverlay:_polyline level:_polyline.level];
    if (_polyline.isVisiable){
        [_mapView updateCachedOverlay:_polyline event:NBVisiableOverlayAdd];
    }
}

- (void)removePolyline {
    [_mapView removeOverlay:_polyline];
    _polylineRenderer = nil;
    [_mapView updateCachedOverlay:_polyline event:NBVisiableOverlayDelete];
}

#pragma mark - <NBMapPolylineOptionsSink>

- (void)setConsumeTapEvents:(BOOL)consume {
    _polyline.tappable = consume;
}

- (void)setVisible:(BOOL)visible {
    if (_polyline.isVisiable){
        [_mapView updateCachedOverlay:_polyline event:NBVisiableOverlayAdd];
    } else{
        [_mapView updateCachedOverlay:_polyline event:NBVisiableOverlayDelete];
    }
    if (_polyline.visible == visible) {
        return;
    }
    if (![self polylineRenderer]) {
        return;
    }
    CGFloat alpha = visible ? 1.0 : 0.0;
    UIColor* newFillColor = [[_polylineRenderer.fillColor copy]colorWithAlphaComponent:alpha];
    UIColor* newStrokeColor = [[_polylineRenderer.strokeColor copy]colorWithAlphaComponent:alpha];
    _polylineRenderer.fillColor = newFillColor;
    _polylineRenderer.strokeColor = newStrokeColor;
    _polyline.fillColor = newFillColor;
    _polyline.strokeColor = newStrokeColor;
    _polyline.visible = visible;
}

- (void)setColor:(UIColor*)color {
    _polyline.strokeColor = color;
    if (![self polylineRenderer]) {
        return;
    }
    [self polylineRenderer].strokeColor = color;
}

- (void)setStrokeWidth:(CGFloat)width {
    _polyline.lineWidth = width;
    if (![self polylineRenderer]) {
        return;
    }
    [self polylineRenderer].lineWidth = width;
}

- (void)setPoints:(NSArray<CLLocation*>*)points {
    NSUInteger count = points.count;
    CLLocationCoordinate2D coords[count];
    for (NSUInteger index = 0; index < count; ++index) {
        CLLocation* location = [points objectAtIndex:index];
        coords[index].latitude = location.coordinate.latitude;
        coords[index].longitude = location.coordinate.longitude;
    }
    [_polyline setPolylineWithCoordinates:coords count:count];
}

///只有在初次添加时才可以设置zIndex
- (void)setZIndex:(int)zIndex {
    if (zIndex >= 1) {
        _polyline.level = MAOverlayLevelAboveLabels;
    } else {
        _polyline.level = MAOverlayLevelAboveRoads;
    }
}

- (void)setNeedsUpdate {
    if(![self polylineRenderer]) {
        return;
    }
    [[self polylineRenderer]setNeedsUpdate];
}
@end

static int ToInt(NSNumber* data) { return [NBJsonConversions toInt:data]; }

static BOOL ToBool(NSNumber* data) { return [NBJsonConversions toBool:data]; }

static NSArray<CLLocation*>* ToPoints(NSArray* data) {
    return [NBJsonConversions toPoints:data];
}

static UIColor* ToColor(NSNumber* data) { return [NBJsonConversions toColor:data]; }

static void InterpretPolylineOptions(NSDictionary* data, id<NBMapPolylineOptionsSink> sink,
                                     NSObject<FlutterPluginRegistrar>* registrar) {
    NSNumber* consumeTapEvents = data[@"consumeTapEvents"];
    if (consumeTapEvents) {
        [sink setConsumeTapEvents:ToBool(consumeTapEvents)];
    }
    
    NSNumber* visible = data[@"visible"];
    if (visible) {
        [sink setVisible:ToBool(visible)];
    }
    
    NSNumber* zIndex = data[@"zIndex"];
    if (zIndex) {
        [sink setZIndex:ToInt(zIndex)];
    }
    
    NSArray* points = data[@"points"];
    if (points) {
        [sink setPoints:ToPoints(points)];
    }
    
    NSNumber* strokeColor = data[@"color"];
    if (strokeColor) {
        [sink setColor:ToColor(strokeColor)];
    }
    
    NSNumber* strokeWidth = data[@"width"];
    if (strokeWidth) {
        [sink setStrokeWidth:ToInt(strokeWidth)];
    }
    [sink setNeedsUpdate];
}

@implementation NBMapPolylinesController {
    NSMutableDictionary* _polylineIdToController;
    FlutterMethodChannel* _methodChannel;
    NSObject<FlutterPluginRegistrar>* _registrar;
    MAMapView* _mapView;
}
- (instancetype)init:(FlutterMethodChannel*)methodChannel
             mapView:(MAMapView*)mapView
           registrar:(NSObject<FlutterPluginRegistrar>*)registrar {
    self = [super init];
    if (self) {
        _methodChannel = methodChannel;
        _mapView = mapView;
        _polylineIdToController = [NSMutableDictionary dictionaryWithCapacity:1];
        _registrar = registrar;
    }
    return self;
}
- (void)addPolylines:(NSArray*)polylinesToAdd {
    for (NSDictionary* polyline in polylinesToAdd) {
        NSArray<CLLocation*>* locations = ToPoints(polyline[@"points"]);
        NSUInteger count = locations.count;
        CLLocationCoordinate2D coords[count];
        for (NSUInteger index = 0; index <count; ++index) {
            CLLocation* location = [locations objectAtIndex:index];
            coords[index] = location.coordinate;
        }
        NSString* polylineId = [NBMapPolylinesController getPolylineId:polyline];
        NBMapPolylineController* controller =
        [[NBMapPolylineController alloc]initPolylineWithCoords:coords count:count polylineId:polylineId mapView:_mapView];
        InterpretPolylineOptions(polyline, controller, _registrar);
        _polylineIdToController[polylineId] = controller;
        [controller addPolyine];
    }
}
- (void)changePolylines:(NSArray*)polylinesToChange {
    for (NSDictionary* polyline in polylinesToChange) {
        NSString* polylineId = [NBMapPolylinesController getPolylineId:polyline];
        NBMapPolylineController* controller = _polylineIdToController[polylineId];
        if (!controller) {
            continue;
        }
        InterpretPolylineOptions(polyline, controller, _registrar);
    }
}
- (void)removePolylineIds:(NSArray*)polylineIdsToRemove {
    for (NSString* polylineId in polylineIdsToRemove) {
        if (!polylineId) {
            continue;
        }
        NBMapPolylineController* controller = _polylineIdToController[polylineId];
        if (!controller) {
            continue;
        }
        [controller removePolyline];
        [_polylineIdToController removeObjectForKey:polylineId];
    }
}
- (void)onPolylineTap:(NSString*)polylineId {
    if (!polylineId) {
        return;
    }
    NBMapPolylineController* controller = _polylineIdToController[polylineId];
    if (!controller) {
        return;
    }
    [_methodChannel invokeMethod:@"polyline#onTap" arguments:@{@"polylineId" : polylineId}];
}
- (bool)hasPolylineWithId:(NSString*)polylineId {
    if (!polylineId) {
        return false;
    }
    return _polylineIdToController[polylineId] != nil;
}

+ (NSString*)getPolylineId:(NSDictionary*)polyline {
    return polyline[@"polylineId"];
}
@end

@implementation NBPolyline

@end
